#include "Dataset.h"
#include "Renderer.h"

/* ------------------------------------------------------------------------ */
Dataset::Dataset(Renderer *r)
{
  char s[50], saux[100];
  FILE *fi;

  R = r;
  dbg = 0;
  printf("Loading the dataset parameters ... ");
  fflush(NULL);

  sprintf(s, "../data/%s.inf", R->DatasetName);
  fi = fopen(s, "rt");
  fgets(saux, 99, fi);
  sscanf(saux, "%d", &nx);
  fgets(saux, 99, fi);
  sscanf(saux, "%d", &ny);
  fgets(saux, 99, fi);
  sscanf(saux, "%d", &nz);
  fgets(saux, 99, fi);
  sscanf(saux, "%f", &sx);
  fgets(saux, 99, fi);
  sscanf(saux, "%f", &sy);
  fgets(saux, 99, fi);
  sscanf(saux, "%f", &sz);
  fclose(fi);

  imaxx = R->maxx - R->minx + 1;
  imaxy = R->maxy - R->miny + 1;
  imaxz = R->maxz - R->minz + 1;

  maxx = (imaxx-1) * sx;
  maxy = (imaxy-1) * sy;
  maxz = (imaxz-1) * sz;

  printf("Done.\n");
  printf("Defining the data volume ... ");
  fflush(NULL);

  Vertices[0] = new Point(0, 0, 0);
  Vertices[1] = new Point(0, 0, maxz);
  Vertices[2] = new Point(0, maxy, 0);
  Vertices[3] = new Point(0, maxy, maxz);
  Vertices[4] = new Point(maxx, 0, 0);
  Vertices[5] = new Point(maxx, 0, maxz);
  Vertices[6] = new Point(maxx, maxy, 0);
  Vertices[7] = new Point(maxx, maxy, maxz);

  Faces[0] = new Face(Vertices[0], Vertices[4], Vertices[6], Vertices[2]);
  Faces[1] = new Face(Vertices[4], Vertices[6], Vertices[7], Vertices[5]);
  Faces[2] = new Face(Vertices[7], Vertices[3], Vertices[1], Vertices[5]);
  Faces[3] = new Face(Vertices[1], Vertices[3], Vertices[2], Vertices[0]);
  Faces[4] = new Face(Vertices[0], Vertices[4], Vertices[5], Vertices[1]);
  Faces[5] = new Face(Vertices[2], Vertices[6], Vertices[7], Vertices[3]);

  Edges[0] = new Edge(Faces[0], Faces[4]);
  Edges[1] = new Edge(Faces[0], Faces[1]);
  Edges[2] = new Edge(Faces[0], Faces[5]);
  Edges[3] = new Edge(Faces[0], Faces[3]);
  Edges[4] = new Edge(Faces[3], Faces[4]);
  Edges[5] = new Edge(Faces[1], Faces[4]);
  Edges[6] = new Edge(Faces[1], Faces[5]);
  Edges[7] = new Edge(Faces[3], Faces[5]);
  Edges[8] = new Edge(Faces[2], Faces[4]);
  Edges[9] = new Edge(Faces[1], Faces[2]);
  Edges[10] = new Edge(Faces[2], Faces[5]);
  Edges[11] = new Edge(Faces[2], Faces[3]);
  printf("Done.\n");

  loadData();
  createMask();
}

/* ------------------------------------------------------------------------ */
float Dataset::getVoxelValue(int x, int y, int z)
{
  if(x < 0 || x > imaxx-1 || y < 0 || y > imaxy-1 || z < 0 || z > imaxz-1)
    return 0.0f;

  return (float)Data[x][y][z]/255.0f;
}

/* ------------------------------------------------------------------------ */
void Dataset::computeDeformation()
{
  int x, y, z;
  int mmaxx, mmaxy, mmaxz, i, level, changes, iaux;
 
  printf("Computing the deformation ... ");
  fflush(NULL);

  mmaxx = imaxx - 1;
  mmaxy = imaxy - 1;
  mmaxz = imaxz - 1;

  for(i=0; i<mmaxx*mmaxy*mmaxz; i++)
    DeformationMask[i] = 0;

  setDeformed(avcx, avcy, avcz);
  Deformation[avcx][avcy][avcz].copy(&avp);
  MinDef.copy(&avp);
  MaxDef.copy(&avp);
  
  level = 1;
  changes = 1;
  while(changes)
  {
    changes = 0;

    for(x=-level; x<level+1; x++)
      for(y=-level; y<level+1; y++)
        for(z=-level; z<level+1; z++)
        {
          if((int)floor(sqrt(x*x+y*y+z*z)) == level)
          {
            iaux = moveVoxel(x, y, z, level);
            if(!changes)
              changes = iaux;
          }
        }

    level++;
    if(level > mmaxx || level > mmaxy || level > mmaxz)
      break;
  }

  mindefx = avcx - level;
  if(mindefx < 0)
    mindefx = 0;
  mindefy = avcy - level;
  if(mindefy < 0)
    mindefy = 0;
  mindefz = avcz - level;
  if(mindefz < 0)
    mindefz = 0;
  maxdefx = avcx + level;
  if(maxdefx > mmaxx)
    maxdefx = mmaxx;
  maxdefy = avcy + level;
  if(maxdefy > mmaxy)
    maxdefy = mmaxy;
  maxdefz = avcz + level;
  if(maxdefz > mmaxz)
    maxdefz = mmaxz;

  printf("Done.\n");
  fflush(NULL);
}


/* ------------------------------------------------------------------------ */
int Dataset::moveVoxel(int x, int y, int z, int level)
{
  int px, py, pz, lx, ly, lz;
  Point np, lp, p, pdir, lavp;
  float eq_dist;
  Ray ray, dir;

  lx = x + avcx;
  ly = y + avcy;
  lz = z + avcz;

  if(lx > imaxx - 1 || ly > imaxy - 1 || lz > imaxz - 1 ||
     lx < 0 || ly < 0 || lz < 0)
    return 0;

  if(isDeformed(lx, ly, lz))
    return 0;

  lp.copy(lx*sx, ly*sy, lz*sz);

  lavp.copy(avcx*sx, avcy*sy, avcz*sz);
  dir.copy(&lp, &lavp, 0);
  dir.getPoint(lp.dist(&lavp)/level, &pdir);
  
  if(lp.dist(&pdir) >= lp.dist(&lavp))
  {
    px = avcx;
    py = avcy;
    pz = avcz;
  }
  else
  {
    px = (int)ceil(pdir.X/sx);
    if(px > imaxx - 1 || px < 0)
    {
      px = (int)floor(pdir.X/sx);
    }
    py = (int)ceil(pdir.Y/sy);
    if(py > imaxy - 1 || py < 0)
    {
      py = (int)floor(pdir.Y/sy);
    }
    pz = (int)ceil(pdir.Z/sy);
    if(pz > imaxz - 1 || pz < 0)
    {
      pz = (int)floor(pdir.Z/sy);
    }
  }

  if(px == lx && py == ly && pz == lz)
  {
    dir.getPoint(2*lp.dist(&lavp)/level, &pdir);
    if(lp.dist(&pdir) >= lp.dist(&lavp))
    {
      px = avcx;
      py = avcy;
      pz = avcz;
    }
    else
    {
      px = (int)ceil(pdir.X/sx);
      if(px > imaxx - 1 || px < 0)
      {
        px = (int)floor(pdir.X/sx);
      }
      py = (int)ceil(pdir.Y/sy);
      if(py > imaxy - 1 || py < 0)
      {
        py = (int)floor(pdir.Y/sy);
      }
      pz = (int)ceil(pdir.Z/sy);
      if(pz > imaxz - 1 || pz < 0)
      {
        pz = (int)floor(pdir.Z/sy);
      }
    }
  }
  
  if(px > imaxx - 1 || py > imaxy - 1 || pz > imaxz - 1 ||
     px < 0 || py < 0 || pz < 0)
    return 0;
  
  if(!isDeformed(px, py, pz))
    return 0;

  eq_dist = sqrt(sx*sx*abs(px-lx) + sy*sy*abs(py-ly) +sz*sz*abs(pz-lz));

  p.copy(px*sx, py*sy, pz*sz);
  p.add(&Deformation[px][py][pz]);

  if(lp.dist(&p) >= 0.5f*eq_dist && lp.dist(&p) <= 1.5f*eq_dist)
    return 0;

  setDeformed(lx, ly, lz);

  ray.copy(&p, &lp, 0);
  if(lp.dist(&p) < 0.5f*eq_dist)
    ray.getPoint(0.5f*eq_dist, &np);
  else
    ray.getPoint(1.5f*eq_dist, &np);

  if(np.greaterThan(&MaxDef))
    MaxDef.copy(&np);
  if(np.lowerThan(&MinDef))
    MinDef.copy(&np);

  lp.sub(&np);
  Deformation[lx][ly][lz].copy(&lp);

  return 1;
}

/* ------------------------------------------------------------------------ */
int Dataset::isDeformed(int x, int y, int z)
{
  unsigned char c;
  int n, ofs;

  n = z*(imaxy-1)*(imaxx-1) + y*(imaxx-1) + x;
  ofs = n % 8;
  n /= 8;
  c = 255;
  c = c << (7-ofs);
  c = c >> 7;
  c = c << ofs;
  c = DeformationMask[n] & c;
  if(c == 0)
    return 0;
  return 1;
}

/* ------------------------------------------------------------------------ */
void Dataset::setDeformed(int x, int y, int z)
{
  unsigned char c;
  int n, ofs;

  n = z*(imaxy-1)*(imaxx-1) + y*(imaxx-1) + x;
  ofs = n % 8;
  n /= 8;
  c = 255;
  c = c << (7-ofs);
  c = c >> 7;
  c = c << ofs;
  DeformationMask[n] = DeformationMask[n] | c;
}

/* ------------------------------------------------------------------------ */
void Dataset::createMask()
{
  int x, y, z, i, ofs;
  unsigned char c;
  int mmaxx, mmaxy, mmaxz;

  printf("Creating the dataset mask ... ");
  fflush(NULL);

  mmaxx = imaxx - 1;
  mmaxy = imaxy - 1;
  mmaxz = imaxz - 1;

  Mask = new unsigned char[mmaxx*mmaxy*mmaxz];
  DeformationMask = new unsigned char[mmaxx*mmaxy*mmaxz];

  for(i=0; i<mmaxx*mmaxy*mmaxz; i++)
    Mask[i] = 0;

  for(z=0; z < mmaxz; z++)
    for(y=0; y < mmaxy; y++)
      for(x=0; x < mmaxx; x++)
      {
        if(Data[x][y][z] != 0 || Data[x][y][z+1] != 0 || 
           Data[x][y+1][z] != 0 || Data[x][y+1][z+1] != 0 || 
           Data[x+1][y][z] != 0 || Data[x+1][y][z+1] != 0 || 
           Data[x+1][y+1][z] != 0 || Data[x+1][y+1][z+1] != 0)
        {
          i = z*mmaxy*mmaxx + y*mmaxx + x;
          ofs = i % 8;
          i /= 8;
          c = 255;
          c = c << (7-ofs);
          c = c >> 7;
          c = c << ofs;
          Mask[i] = Mask[i] | c;
        }
      }
  printf("Done.\n");
  fflush(NULL);
}

/* ------------------------------------------------------------------------ */
void Dataset::loadData()
{
  FILE *f;
  int x, y, z;
  unsigned char uaux;
  int pct;
  char s[50];

  printf("Loading the dataset ...  0%%");
  fflush(NULL);

  Data = new (unsigned char**)[imaxx];
  Deformation = new (Point**)[imaxx];
  for(x=0; x < imaxx; x++)
  {
    Data[x] = new (unsigned char*)[imaxy];
    Deformation[x] = new (Point*)[imaxy];
    for(y=0; y < imaxy; y++)
    {
      Data[x][y] = new unsigned char[imaxz];
      Deformation[x][y] = new Point[imaxz];
    }
  }

  sprintf(s, "../data/%s.dat", R->DatasetName);
  f = fopen(s, "rb");

  x = 0;
  y = 0;
  z = 0;
  while(!feof(f))
  {
    fread(&uaux, 1, 1, f);

    if(x >= R->minx && x <= R->maxx &&
       y >= R->miny && y <= R->maxy &&
       z >= R->minz && z <= R->maxz)
      Data[x-R->minx][y-R->miny][z-R->minz] = uaux;

    x = (x+1)%nx;
    if(x == 0)
    {
      y = (y+1)%ny;
      if(y == 0)
      {
        z = (z+1)%nz;

        pct = (int)floor((float)(z+1)/(float)nz*100.0f);
        if(pct < 10)
          printf("\b\b");
        else
          printf("\b\b\b");
        printf("%d%%", pct);
        fflush(NULL);
      }
    }
  }

  fclose(f);
  printf("\b\b\b\bDone.\n");
  fflush(NULL);
}


/* ------------------------------------------------------------------------ */
Dataset::~Dataset()
{
  int x, y;

  for(x=0; x<imaxx; x++)
  {
    for(y=0; y<imaxy; y++)
      delete []Data[x][y];
    delete []Data[x];
  }
  delete []Data;
      
}

/* ------------------------------------------------------------------------ */
int Dataset::getPointNormal(Point *p, Point *n)
{
  Hood hd;
  Point paux;

  if(!getHood(p, &hd))
    return 0;

  n->copy(0.0f, 0.0f, 0.0f);
  getVoxelNormal(hd.minx, hd.miny, hd.minz, &paux);
  n->add(&paux);
  getVoxelNormal(hd.minx, hd.miny, hd.maxz, &paux);
  n->add(&paux);
  getVoxelNormal(hd.minx, hd.maxy, hd.minz, &paux);
  n->add(&paux);
  getVoxelNormal(hd.minx, hd.maxy, hd.maxz, &paux);
  n->add(&paux);
  getVoxelNormal(hd.maxx, hd.miny, hd.minz, &paux);
  n->add(&paux);
  getVoxelNormal(hd.maxx, hd.miny, hd.maxz, &paux);
  n->add(&paux);
  getVoxelNormal(hd.maxx, hd.maxy, hd.minz, &paux);
  n->add(&paux);
  getVoxelNormal(hd.maxx, hd.maxy, hd.maxz, &paux);
  n->add(&paux);
  n->normalize();

  return 1;
}

/* ------------------------------------------------------------------------ */
void Dataset::getVoxelNormal(int x, int y, int z, Point *n)
{
  Point xp, xm, yp, ym, zp, zm, v;
  float cxp, cxm, cyp, cym, czp, czm;

  getPosition(x, y, z, &v);
  getPosition(x+1, y, z, &xp);
  getPosition(x-1, y, z, &xm);
  getPosition(x, y+1, z, &yp);
  getPosition(x, y-1, z, &ym);
  getPosition(x, y, z+1, &zp);
  getPosition(x, y, z-1, &zm);

  xp.sub(&v);
  xm.sub(&v);
  yp.sub(&v);
  ym.sub(&v);
  zp.sub(&v);
  zm.sub(&v);

  if(xp.X == 0)
    cxp = 0.0f;
  else
    cxp = sx/xp.X;

  if(xm.X == 0)
    cxm = 0.0f;
  else
    cxm = -sx/xm.X;

  if(yp.Y == 0)
    cyp = 0.0f;
  else
    cyp = sy/yp.Y;

  if(ym.Y == 0)
    cym = 0.0f;
  else
    cym = -sy/ym.Y;

  if(zp.Z == 0)
    czp = 0.0f;
  else
    czp = sz/zp.Z;

  if(zm.Z == 0)
    czm = 0.0f;
  else
    czm = -sz/zm.Z;

  n->copy(getVoxelValue(x+1, y, z)*cxp - getVoxelValue(x-1, y, z)*cxm,
          getVoxelValue(x, y+1, z)*cyp - getVoxelValue(x, y-1, z)*cym,
          getVoxelValue(x, y, z+1)*czp - getVoxelValue(x, y, z-1)*czm);

  if(n->X == 0.0f && n->Y == 0.0f && n->Z == 0.0f)
  {
    n->X = 1.0f;
    n->Y = 1.0f;
    n->Z = 1.0f;
  }
  n->normalize();
}

/* ------------------------------------------------------------------------ */
int Dataset::getHood(Point *p, Hood *h)
{
  int mx, my, mz, Mx, My, Mz;
  float x, y, z;

  x = p->X;
  y = p->Y;
  z = p->Z;

  if(x < 0.0 || y < 0.0 || z < 0.0 ||x > maxx || y > maxy || z > maxz)
    return 0;

  mx = (int)floor(x/sx);
  Mx = (int)ceil(x/sx);
  my = (int)floor(y/sy);
  My = (int)ceil(y/sy);
  mz = (int)floor(z/sz);
  Mz = (int)ceil(z/sz);

  if(Mx > R->maxx)
    Mx = R->maxx;
  if(My > R->maxy)
    My = R->maxy;
  if(Mz > R->maxz)
    Mz = R->maxz;

  h->copy(mx, my, mz, Mx, My, Mz);
  return 1;
}

/* ------------------------------------------------------------------------ */
int Dataset::isInside(Point *p)
{
  if(p->X < 0.0 || p->Y < 0.0 || p->Z < 0.0 ||
     p->X > maxx || p->Y > maxy || p->Z > maxz)
    return 0;
  return 1;
}

/* ------------------------------------------------------------------------ */
void Dataset::getPosition(int x, int y, int z, Point *p)
{
  if(x < 0 || x > imaxx-1 || y < 0 || y > imaxy-1 || z < 0 || z > imaxz-1)
  {
    p->init();
    return;
  }

  p->X = x*sx;
  p->Y = y*sy;
  p->Z = z*sz;
  if(isDeformed(x, y, z))
    p->add(&Deformation[x][y][z]);
}

/* ------------------------------------------------------------------------ */
int Dataset::goodDeformedVoxel(Point *v, Point *p)
{
  Point dg, lp, lv;
  
  dg.copy(avcx*sx, avcy*sy, avcz*sz);
  lp.copy(p);
  lv.copy(v);
  dg.sub(&lp);
  lv.sub(&lp);
  if(dg.X*lv.X + dg.Y*lv.Y + dg.Z*lv.Z < 0)
    return 0;
  return 1;
}

/* ------------------------------------------------------------------------ */
float Dataset::getValue(Point *p)
{
  float v[4], x, y, z, vf, defretval;
  int mx, my, mz, Mx, My, Mz, n, ofs, ix, iy, iz;
  float tx, ty, tz, Tx, Ty, Tz, mindist;
  unsigned char mmm, mmM, mMm, mMM, Mmm, MmM, MMm, MMM;
  unsigned char c;
  Hood hd;
  Point paux;
  
  if(!getHood(p, &hd))
    return 0;

  mx = hd.minx;
  Mx = hd.maxx;
  my = hd.miny;
  My = hd.maxy;
  mz = hd.minz;
  Mz = hd.maxz;

  if(p->greaterThan(&MinDef) && p->lowerThan(&MaxDef) &&
     (isDeformed(mx, my, mz) ||
      isDeformed(mx, my, Mz) ||
      isDeformed(mx, My, mz) ||
      isDeformed(mx, My, Mz) ||
      isDeformed(Mx, my, mz) ||
      isDeformed(Mx, my, Mz) ||
      isDeformed(Mx, My, mz) ||
      isDeformed(Mx, My, Mz)))
  {
    defretval = 0.0f;
    mindist = 2*R->DSize;

    for(ix=mindefx; ix<maxdefx+1; ix++)
      for(iy=mindefy; iy<maxdefy+1; iy++)
        for(iz=mindefz; iz<maxdefz+1; iz++)
        {
          getPosition(ix, iy, iz, &paux);
          if(mindist > paux.dist(p) && goodDeformedVoxel(&paux, p))
          {
            mindist= paux.dist(p);
            defretval = Data[ix][iy][iz];
          }
        }

    getPosition(mx, my, mz, &paux);
    if(mindist> paux.dist(p) && goodDeformedVoxel(&paux, p))
    {
      mindist= paux.dist(p);
      defretval = Data[mx][my][mz];
    }
    getPosition(mx, my, Mz, &paux);
    if(mindist> paux.dist(p) && goodDeformedVoxel(&paux, p))
    {
      mindist= paux.dist(p);
      defretval = Data[mx][my][Mz];
    }
    getPosition(mx, My, mz, &paux);
    if(mindist> paux.dist(p) && goodDeformedVoxel(&paux, p))
    {
      mindist= paux.dist(p);
      defretval = Data[mx][My][mz];
    }
    getPosition(mx, My, Mz, &paux);
    if(mindist> paux.dist(p) && goodDeformedVoxel(&paux, p))
    {
      mindist= paux.dist(p);
      defretval = Data[mx][My][Mz];
    }
    getPosition(Mx, my, mz, &paux);
    if(mindist> paux.dist(p) && goodDeformedVoxel(&paux, p))
    {
      mindist= paux.dist(p);
      defretval = Data[Mx][my][mz];
    }
    getPosition(Mx, my, Mz, &paux);
    if(mindist> paux.dist(p) && goodDeformedVoxel(&paux, p))
    {
      mindist= paux.dist(p);
      defretval = Data[Mx][my][Mz];
    }
    getPosition(Mx, My, mz, &paux);
    if(mindist> paux.dist(p) && goodDeformedVoxel(&paux, p))
    {
      mindist= paux.dist(p);
      defretval = Data[Mx][My][mz];
    }
    getPosition(Mx, My, Mz, &paux);
    if(mindist> paux.dist(p) && goodDeformedVoxel(&paux, p))
    {
      mindist= paux.dist(p);
      defretval = Data[Mx][My][Mz];
    }
    return defretval;
  }

  x = p->X/sx;
  y = p->Y/sy;
  z = p->Z/sz;

  tx = x - mx;
  Tx = Mx - x;
  ty = y - my;
  Ty = My - y;
  tz = z - mz;
  Tz = Mz - z;

  mmm = Data[mx][my][mz];
  mmM = Data[mx][my][Mz];
  mMm = Data[mx][My][mz];
  mMM = Data[mx][My][Mz];
  Mmm = Data[Mx][my][mz];
  MmM = Data[Mx][my][Mz];
  MMm = Data[Mx][My][mz];
  MMM = Data[Mx][My][Mz];

  if(mx != Mx && my != My && mz != Mz)
  {
    n = mz*(imaxy-1)*(imaxx-1) + my*(imaxx-1) + mx;
    ofs = n % 8;
    n /= 8;
    c = 255;
    c = c << (7-ofs);
    c = c >> 7;
    c = c << ofs;
    c = Mask[n] & c;
    if(c == 0)
      return 0.0f;
  }

  if(mx == Mx && my == My && mz == Mz)
    return mmm/255.0f;

  if(mx == Mx && my == My)
  {
    vf = mmm*tz + mmM*Tz;
    if(vf < 0.0f)
      vf = 0.0f;
    if(vf > 255.0f)
      vf = 255.0f;
    return vf/255.0f;
  }

  if(mx == Mx && mz == Mz)
  {
    vf = mmm*ty + mMm*Ty;
    if(vf < 0.0f)
      vf = 0.0f;
    if(vf > 255.0f)
      vf = 255.0f;
    return vf/255.0f;
  }

  if(my == My && mz == Mz)
  {
    vf = mmm*tx + Mmm*Tx;
    if(vf < 0.0f)
      vf = 0.0f;
    if(vf > 255.0f)
      vf = 255.0f;
    return vf/255.0f;
  }

  if(mx == Mx)
  {
    v[0] = mmm*ty + mMm*Ty;
    v[1] = mmM*ty + mMM*Ty;
    vf = v[0]*tz + v[1]*Tz;
    if(vf < 0.0f)
      vf = 0.0f;
    if(vf > 255.0f)
      vf = 255.0f;
    return vf/255.0f;
  }

  if(my == My)
  {
    v[0] = mmm*tx + Mmm*Tx;
    v[1] = mmM*tx + MmM*Tx;
    vf = v[0]*tz + v[1]*Tz;
    if(vf < 0.0f)
      vf = 0.0f;
    if(vf > 255.0f)
      vf = 255.0f;
    return vf/255.0f;
  }

  if(mz == Mz)
  {
    v[0] = mmm*tx + Mmm*Tx;
    v[1] = mMm*tx + MMm*Tx;
    vf = v[0]*ty + v[1]*Ty;
    if(vf < 0.0f)
      vf = 0.0f;
    if(vf > 255.0f)
      vf = 255.0f;
    return vf/255.0f;
  }

  v[0] = mmm*tx +Mmm*Tx;
  v[1] = mMm*tx + MMm*Tx;
  v[0] = v[0]*ty + v[1]*Ty/255.0f;
  
  v[2] = mmM*tx +MmM*Tx;
  v[3] = mMM*tx + MMM*Tx;
  v[2] = v[2]*ty + v[3]*Ty;
  
  vf = v[0]*tz + v[2]*Tz;
  if(vf < 0.0f)
    vf = 0.0f;
  if(vf > 255.0f)
    vf = 255.0f;
  return vf/255.0f;
}

/* ------------------------------------------------------------------------ */
int Dataset::intersect(Ray *r)
{
  int i;
  float t;
  
  for(i=0; i<6; i++)
    if(Faces[i]->getIntersection(r, t))
      return 1;

  return 0;
}

/* ------------------------------------------------------------------------ */
int Dataset::onFace(Point *p)
{
  int i;
  float trhl;

  trhl = R->DSize/R->WSize;
  
  for(i=0; i<6; i++)
    if(Faces[i]->near(p, trhl))
      return 1;

  return 0;
}

/* ------------------------------------------------------------------------ */
int Dataset::getExtremeIntersections(Ray *ray, float &first, float &last)
{
  int aux;
  float t, min, max, x, y, z;
  Point p;
  
  min = -1.0f;
  max = -1.0f;
  aux = 0;

  if(ray->Dx != 0)
  {
    t = - ray->X0/ray->Dx;
    y = ray->Dy*t+ray->Y0;
    z = ray->Dz*t+ray->Z0;
    if(y >= 0.0f && y <= maxy && z >= 0.0f && z <= maxz)
    {
      if(min > t || min < 0.0f)
        min = t;
      if(max < t || max < 0.0f)
        max = t;
      aux = 1;
    }
  }
  if(ray->Dx != 0)
  {
    t = (maxx - ray->X0)/ray->Dx;
    y = ray->Dy*t+ray->Y0;
    z = ray->Dz*t+ray->Z0;
    if(y >= 0.0f && y <= maxy && z >= 0.0f && z <= maxz)
    {
      if(min > t || min < 0.0f)
        min = t;
      if(max < t || max < 0.0f)
        max = t;
      aux = 1;
    }
  }

  if(ray->Dy != 0)
  {
    t = - ray->Y0/ray->Dy;
    x = ray->Dx*t+ray->X0;
    z = ray->Dz*t+ray->Z0;
    if(x >= 0.0f && x <= maxx && z >= 0.0f && z <= maxz)
    {
      if(min > t || min < 0.0f)
        min = t;
      if(max < t || max < 0.0f)
        max = t;
      aux = 1;
    }
  }
  if(ray->Dy != 0)
  {
    t = (maxy- ray->Y0)/ray->Dy;
    x = ray->Dx*t+ray->X0;
    z = ray->Dz*t+ray->Z0;
    if(x >= 0.0f && x <= maxx && z >= 0.0f && z <= maxz)
    {
      if(min > t || min < 0.0f)
        min = t;
      if(max < t || max < 0.0f)
        max = t;
      aux = 1;
    }
  }

  if(ray->Dz != 0)
  {
    t = - ray->Z0/ray->Dz;
    x = ray->Dx*t+ray->X0;
    y = ray->Dy*t+ray->Y0;
    if(x >= 0.0f && x <= maxx && y >= 0.0f && y <= maxy)
    {
      if(min > t || min < 0.0f)
        min = t;
      if(max < t || max < 0.0f)
        max = t;
      aux = 1;
    }
  }
  if(ray->Dz != 0)
  {
    t = (maxz - ray->Z0)/ray->Dz;
    x = ray->Dx*t+ray->X0;
    y = ray->Dy*t+ray->Y0;
    if(x >= 0.0f && x <= maxx && y >= 0.0f && y <= maxy)
    {
      if(min > t || min < 0.0f)
        min = t;
      if(max < t || max < 0.0f)
        max = t;
      aux = 1;
    }
  }

  if(aux)
  {
    ray->getPoint(0.0f, &p);
    if(isInside(&p))
      first = 0.00001f;
    else
      first = min;
    last = max;
    return 1;
  }
  else
    return 0;
}

/* ------------------------------------------------------------------------ */
int Dataset::isOnEdge(Point *p)
{
  float xi, yi, zi, trhl;
  int i;

  xi = p->X;
  yi = p->Y;
  zi = p->Z;

  trhl = R->DSize/R->WSize;

  if(xi >= 0 && xi <= maxx+trhl && 
     yi >= 0 && yi <= maxy+trhl && 
     zi >= 0 && zi <= maxz+trhl)
    for(i=0; i<12; i++)
      if(Edges[i]->near(p, trhl))
        return 1;
  return 0;
}

